;; Generated package description from rcirc-sqlite.el  -*- no-byte-compile: t -*-
(define-package "rcirc-sqlite" "0.1.3.0.20240213.113754" "rcirc logging in SQLite" '((emacs "30.0")) :commit "7549f95511be4d80610714d5b957dc05aa8ee7fc" :authors '(("Matto Fransen" . "matto@matto.nl")) :maintainer '("Matto Fransen" . "matto@matto.nl") :keywords '("comm") :url "https://codeberg.org/mattof/rcirc-sqlite")
